/*=========================*/
/* Fellow Emulator         */
/* Sound emulation         */
/* Device-independent part */
/* Petter Schau 1996/1997  */
/*=========================*/

#include <stdio.h>
#include "defs.h"
#include "memory.h"

#ifndef SOUND_C
#include "sounda.h"
#endif

/*===============================*/
/* Sound emulation configuration */
/*===============================*/

/* Sound-modes available, set to TRUE or FALSE by the device-dependent driver

   0  - 44100 stereo 16 bit
   1  - 31300 stereo 16 bit
   2  - 22050 stereo 16 bit
   3  - 15650 stereo 16 bit
   4  - 44100 mono 16 bit
   5  - 31300 mono 16 bit
   6  - 22050 mono 16 bit
   7  - 15650 mono 16 bit
   8  - 44100 stereo 8 bit
   9  - 31300 stereo 8 bit
   10 - 22050 stereo 8 bit
   11 - 15650 stereo 8 bit
   12 - 44100 mono 8 bit
   13 - 31300 mono 8 bit
   14 - 22050 mono 8 bit
   15 - 15650 mono 8 bit */

ULO config_soundmodes[16];

/* The active sound quality, 0 - 16, as described above */

ULO config_sound_mode;

/* Startup flag, don't set sound quality to default when this is TRUE */

ULO configaudiomodesetbylog;

/* Callback routines for setup of a specific sound-mode */
/* Called when the emulator (user) changes the sound-mode */
/* The soundcard driver can then do any adjustments */

setup_soundmode_routine config_sound_setuproutines[16];

/* Current sound emulation setting */
/* 0 - none, 1 - normal, 2 - "contiguous", 3 - only emulate */

ULO config_sound;      

/* Set TRUE by a device-dependent driver */

ULO config_soundcard_found;

/* Callback routines for adding samples to souncard specific buffers */
/* Set by soundcard driver, correct routine moved to audio_putinbuffer */
/* when soundmode is changed. */

add_sample_to_buffer_routine config_sound_bufferroutines[16];

/*=================================================*/
/* Callback pointers used by emulation at run-time */
/*=================================================*/

/* Frequency handler called from eol */

audiofreqhandlerfunc audio_handler;

/* Called by frequency handler (card specific, set by card-driver) */

add_sample_to_buffer_routine audio_putinbuffer;

/* Callback to play buffer, set by device-dependent driver */

playbuffer_routine sound_playbuffer;

/* Callbacks called before and after emulation is started/stopped */
/* Set by device-dependent driver */

sound_before_emu_routine sound_before_emu;
sound_after_emu_routine  sound_after_emu;

/*===============*/
/* Run-time data */
/*===============*/

/* Temorary buffer for samples generated by sound-emulation, 16-bit signed */

UWO left[8],right[8];

/* Used in 22050/44100 to decide samples */

ULO audiocounter;

/* Used for skipping samples in 22050 */

ULO audioodd;

/* Count 5 frames, and then play */

ULO sound_framecounter;

/* Used by the "contiguous sound" option */

ULO audio_hblanks;
ULO audio_hblanks_thisbuffer;

/*=================*/
/* Audio-registers */
/*=================*/

ULO audpt[4];             /* Sample-DMA pointer */
ULO audlen[4];            /* Length */
ULO audper[4];            /* Used directly, NOTE: translated value */ 
ULO audvol[4];            /* Volume, possibly not reloaded by state-machine */
ULO auddat[4];            /* Last data word read by DMA */

/*==========================================*/
/* Internal variables used by state-machine */
/*==========================================*/

ULO audlenw[4];             /* Length counter */
ULO audpercounter[4];       /* Period counter */
ULO auddatw[4];             /* Sample currently output, 16-bit signed */
audiostatefunc audstate[4]; /* Current state for the channel */
ULO audvolw[4];             /* Current volume, reloaded at some points */
ULO audptw[4];              /* Current dma-pointer, reloaded at some points */ 

/*====================*/
/* Translation tables */
/*====================*/

ULO periodtable[65536];
WOR volumes[256][64];
ULO audioirqmask[4] = {0x0080,0x0100,0x0200,0x0400};
ULO audiodmaconmask[4] = {0x1,0x2,0x4,0x8};

void audiostate0(int);
void audiostate1(int);
void audiostate2(int);
void audiostate3(int);
void audiostate5(int);

void sound_callback_dummy(ULO newsamples) {};
ULO sound_callback_initmode_dummy(ULO mode) {return 31300;};

#ifdef SOUND_C

/*======================*/
/* Register write stubs */
/*======================*/

#define GET_CHANNEL_NUMBER(address) (((address>>4) & 7) - 2)

/* AUDXPTH - $dff0a0,b0,c0,d0 */

#pragma aux waudXpth parm [ecx] [edx];
void waudXpth(int address, int data) {
  push_eax();
  *(((UWO *) &audpt[GET_CHANNEL_NUMBER(address)]) + 1) = data & 0x1f;
  pop_eax();
}

/* AUDXPTL - $dff0a2,b2,c2,d2 */

#pragma aux waudXptl parm [ecx] [edx];
void waudXptl(int address, int data) {
  push_eax();
  *((UWO *) &audpt[GET_CHANNEL_NUMBER(address)]) = data & 0xfffe;
  pop_eax();
}

/* AUDXLEN - $dff0a4,b4,c4,d4 */

#pragma aux waudXlen parm [ecx] [edx];
void waudXlen(int address, int data) {
  push_eax();
  audlen[GET_CHANNEL_NUMBER(address)] = data & 0xffff;
  pop_eax();
}

/* AUDXPER - $dff0a6,b6,c6,d6 */

#pragma aux waudXper parm [ecx] [edx];
void waudXper(int address, int data) {
  push_eax();
  audper[GET_CHANNEL_NUMBER(address)] = periodtable[data & 0xffff];
  pop_eax();
}

/* AUDXVOL - $dff0a8,b8,c8,d8 */

#pragma aux waudXvol parm [ecx] [edx];
void waudXvol(int address, int data) {
  push_eax();
  if ((data & 0xff00) && !(data & 0xff)) data >>= 8; 
  if (data & 64) data = 63;
  audvol[GET_CHANNEL_NUMBER(address)] = (data & 63);
  pop_eax();
}

/*=====================*/
/* Audio state machine */
/*=====================*/

/* Inline intreq */
/* Can be replaced by the slower wriw(address,data) */

void wintreq_inline(ULO data);
#pragma aux wintreq_inline parm [edx] = "call   wintreq";

/* State 0 */
/* Never called when dma is off */

void audiostate0(int channel) {
  audlenw[channel] = audlen[channel]; /* State-change 0 to 1 */
  audptw[channel] = audpt[channel];
  audpercounter[channel] = 0;
  audstate[channel] = audiostate1;
}

/* State 1 */

void audiostate1(int channel) {
  if (audlenw[channel] != 1) audlenw[channel]--;   /* State-change 1 to 5 */
  audstate[channel] = audiostate5;
  wintreq_inline(0x8000 | audioirqmask[channel]);
}

/* State 2 */

void audiostate2(int channel) {
  if (audpercounter[channel] >= 0x10000) {      /* State-change 2 to 3 */
    audpercounter[channel] = audpercounter[channel] - 0x10000 + audper[channel];
    audvolw[channel] = audvol[channel];
    auddatw[channel] = volumes[*((char *) &auddat[channel])][audvolw[channel]];
    audstate[channel] = audiostate3;
    }
  else audpercounter[channel] += audper[channel]; /* State-change 2 to 2 */
}

/* State 3 */

void audiostate3(int channel) {
  if (audpercounter[channel] >= 0x10000) {     /* State-change 3 to 2 */
    audpercounter[channel] = audpercounter[channel] - 0x10000 + audper[channel];
    auddatw[channel] = volumes[*(((char *) &auddat[channel]) + 1)][audvolw[channel]];
    auddat[channel] = (((ULO) cmem[audptw[channel]])<<8) | cmem[audptw[channel] + 1];
    audptw[channel] = (audptw[channel] + 2) & 0x1ffffe;
    audstate[channel] = audiostate2;
    if (audlenw[channel] == 1) {
      audlenw[channel] = audlen[channel];
      audptw[channel] = audpt[channel];
      wintreq_inline(0x8000 | audioirqmask[channel]);
      }
    else audlenw[channel]--;
  }
  else audpercounter[channel] += audper[channel]; /* State-change 3 to 3 */
}    			

/* State 5 */

void audiostate5(int channel) {
  audvolw[channel] = audvol[channel];
  audpercounter[channel] = 0;
  auddatw[channel] = volumes[*(((char *) &auddat[channel]) + 1)][audvolw[channel]];
  auddat[channel] = (((ULO) cmem[audptw[channel]])<<8) | cmem[audptw[channel] + 1];
  audptw[channel] = (audptw[channel] + 2) & 0x1ffffe;
  audstate[channel] = audiostate2;
  if (audlenw[channel] == 1) {
    audlenw[channel] = audlen[channel];
    audptw[channel] = audpt[channel];
    wintreq_inline(0x8000 | audioirqmask[channel]);
    }
  else audlenw[channel]--;
}

/* Called by DMACON write when dma for a channel is turned off */
/* This is why audiostates do not check DMA on/off */

#pragma aux audio_kill_channel parm [ecx];
void audio_kill_channel(int channel) {
  push_eax();
  auddatw[channel] = 0;
  audstate[channel] = audiostate0;
  pop_eax();
}

/* Frequency handler for 44100 hz */

void audio_frequency_handler_44100(void) {
  push_eax();
  *((ULO *) left) = 0;
  *((ULO *) right) = 0;
  if (dmacon & 1) {
    audstate[0](0);
    left[0] = auddatw[0];
    audstate[0](0);
    left[1] = auddatw[0];
    }
  if (dmacon & 8) {
    audstate[3](3);
    left[0] += auddatw[3];
    audstate[3](3);
    left[1] += auddatw[3];
    }
  if (dmacon & 2) {
    audstate[1](1);
    right[0] = auddatw[1];
    audstate[1](1);
    right[1] = auddatw[1];
    }
  if (dmacon & 4) {
    audstate[2](2);
    right[0] += auddatw[2];
    audstate[2](2);
    right[1] += auddatw[2];
    }
  audio_putinbuffer(2);
  pop_eax();
}
/* Frequency handler for 31300 hz */

void audio_frequency_handler_31300(void) {
  push_eax();
  *((ULO *) left) = 0;
  *((ULO *) right) = 0;
  if (dmacon & 1) {
    audstate[0](0);
    left[0] = auddatw[0];
    audstate[0](0);
    left[1] = auddatw[0];
    }
  if (dmacon & 8) {
    audstate[3](3);
    left[0] += auddatw[3];
    audstate[3](3);
    left[1] += auddatw[3];
    }
  if (dmacon & 2) {
    audstate[1](1);
    right[0] = auddatw[1];
    audstate[1](1);
    right[1] = auddatw[1];
    }
  if (dmacon & 4) {
    audstate[2](2);
    right[0] += auddatw[2];
    audstate[2](2);
    right[1] += auddatw[2];
    }
  audio_putinbuffer(2);
  pop_eax();
}

/* Frequency handler for 22050 hz */

void audio_frequency_handler_22050(void) {
  push_eax();
  *((ULO *) left) = 0;
  *((ULO *) right) = 0;
  if (dmacon & 1) {
    audstate[0](0);
    left[0] = auddatw[0];
    audstate[0](0);
    left[1] = auddatw[0];
    }
  if (dmacon & 8) {
    audstate[3](3);
    left[0] += auddatw[3];
    audstate[3](3);
    left[1] += auddatw[3];
    }
  if (dmacon & 2) {
    audstate[1](1);
    right[0] = auddatw[1];
    audstate[1](1);
    right[1] = auddatw[1];
    }
  if (dmacon & 4) {
    audstate[2](2);
    right[0] += auddatw[2];
    audstate[2](2);
    right[1] += auddatw[2];
    }
  audio_putinbuffer(2);
  pop_eax();
}

/* Frequency handler for 15650 hz */

void audio_frequency_handler_15650(void) {
  push_eax();
  *((ULO *) left) = 0;
  *((ULO *) right) = 0;
  if (dmacon & 1) {
    audstate[0](0);
    left[0] = auddatw[0];
    audstate[0](0);
    left[1] = auddatw[0];
    }
  if (dmacon & 8) {
    audstate[3](3);
    left[0] += auddatw[3];
    audstate[3](3);
    left[1] += auddatw[3];
    }
  if (dmacon & 2) {
    audstate[1](1);
    right[0] = auddatw[1];
    audstate[1](1);
    right[1] = auddatw[1];
    }
  if (dmacon & 4) {
    audstate[2](2);
    right[0] += auddatw[2];
    audstate[2](2);
    right[1] += auddatw[2];
    }
  audio_putinbuffer(2);
  pop_eax();
}

#endif




/* Must be called every time sound quality is changed */

void initvolumetable(void) {
  int i,s,j;

  if (config_sound_mode & 4) s = 1;    /* Mono */
  else s = 2;                          /* Stereo */

  for (i = -128; i < 128; i++) 
    for (j = 0; j < 64; j++) {
      if (j == 0) volumes[(i&0xff)][j] = 0;
      else volumes[(i&0xff)][j] = (i*j*s)/2;
      }
}

/* Must be called every time sound_quality is changed */
/* baseperiod is the output frequency */

void initperiodtable(ULO baseperiod) {
  double j;
  int i;

  periodtable[0] = 0x10000;
  for (i = 1; i < 65536; i++) {
    j = 3568200/i;                  /* Sample rate */
    periodtable[i] = (ULO) ((j*65536)/baseperiod);
    if (periodtable[i] > 0x10000) periodtable[i] = 0x10000;
    }
}

/* Called whenever a new sound quality is selected */
/* Call the sound card init routine for this quality */

void sound_initmode(void) {
  int tmp = config_sound_mode;

  audiocounter = 0;
  if (config_sound == 3) {
    initperiodtable(31300);
    config_sound_mode = 1;
    initvolumetable();
    config_sound_mode = tmp;
    audio_handler = audio_frequency_handler_15650;
    audio_putinbuffer = sound_callback_dummy;
    sound_playbuffer = sound_callback_dummy;
    sound_before_emu = sound_callback_dummy;
    sound_after_emu = sound_callback_dummy;
    }
  else if (config_sound > 0) {
    initperiodtable((*config_sound_setuproutines[config_sound_mode])());
    initvolumetable();
    switch ((config_sound_mode & 0x3)) {
      case 0: audio_handler = audio_frequency_handler_44100;
              break;
      case 1: audio_handler = audio_frequency_handler_31300;
              break;
      case 2: audio_handler = audio_frequency_handler_22050;
              break;
      case 3: audio_handler = audio_frequency_handler_15650;
              break;
      }
    audio_putinbuffer = config_sound_bufferroutines[config_sound_mode];
    }
  else {
    audio_handler = sound_callback_dummy;
    audio_putinbuffer = sound_callback_dummy;
    sound_playbuffer = sound_callback_dummy;
    sound_before_emu = sound_callback_dummy;
    sound_after_emu = sound_callback_dummy;
    }
}

void sound_clear_callbacks(void) {
  int i;

  for (i = 0; i < 16; i++) {
    config_sound_bufferroutines[i] = (add_sample_to_buffer_routine)sound_callback_dummy;
    config_sound_setuproutines[i] = (setup_soundmode_routine)sound_callback_initmode_dummy;
    config_soundmodes[i] = FALSE;
    }
  sound_playbuffer = sound_callback_dummy;
  sound_before_emu = sound_callback_dummy;
  sound_after_emu = sound_callback_dummy;
}

/* Called every time we do a hard-reset */

void sound_reset(void) {
  int i;

  for (i = 0; i < 4; i++) {
    audpt[i] = 0;
    audptw[i] = 0;
    audlen[i] = 2;
    audper[i] = 0;
    audvol[i] = 0;
    audpercounter[i] = 0;
    auddat[i] = 0;
    auddatw[i] = 0;
    audlenw[i] = 2;
    audstate[i] = audiostate0;
    audvolw[i] = 0;
    iowrite[(0xa0>>1) + (0x8*i)] = waudXpth;
    iowrite[(0xa2>>1) + (0x8*i)] = waudXptl;
    iowrite[(0xa4>>1) + (0x8*i)] = waudXlen;
    iowrite[(0xa6>>1) + (0x8*i)] = waudXper;
    iowrite[(0xa8>>1) + (0x8*i)] = waudXvol;
    }
#ifdef SOUND_C       
  audio_hblanks = (ULO) audio_kill_channel;
#endif
  audio_hblanks = 0;
  audio_hblanks_thisbuffer = 0;
}

/* Called once, and before any soundcard setup */

void soundinit(void) {
  sound_reset();
  sound_clear_callbacks();
  config_soundcard_found = FALSE;
  audioodd = 0;
}

